#!/usr/bin/env python

"""
Small utility that parses Netscape bookmarks.
"""

from xml.sax import handler
import bookmark
import string


class NetscapeHandler(handler.ContentHandler):
    """ SAX handler for Netscape bookmarks """

    def __init__(self):
        self.bms = bookmark.Bookmarks()
        self.cur_elem = None
        self.added    = None
        self.href     = None
        self.visited  = None
        self.modified = None

    def startElement(self, name, attrs):
        name = string.lower( name )
        d = {}
        for key, value in attrs.items():
            d[ string.lower(key) ] = value
        if name=="h3":
            self.cur_elem="h3"
            if d.has_key('add_date'): self.added=d["add_date"]
            else: self.added = ""

        elif name=="a":
            self.cur_elem="a"
            if d.has_key('add_date'): self.added=d["add_date"]
            else: self.added = None
            if d.has_key('last_visit'): self.visited=d["last_visit"]
            else: self.visited = None
            if d.has_key('last_modified'): self.modified=d["last_modified"]
            else: self.modified = None
            
            self.url=d["href"]
        elif name=='title':  # Could equally use h1 element
            self.cur_elem = 'title'
            self.bms.owner = ""
            
    def characters(self,data):
##        print 'char', self.cur_elem, data[start:start+length]
        if self.cur_elem=="h3":
            self.bms.add_folder(data,None)
        elif self.cur_elem=="a":
            self.bms.add_bookmark(data, 
                                  added = self.added,
                                  visited = self.visited,
                                  modified = self.modified,
                                  href = self.url)
        elif self.cur_elem=="title":
            self.bms.owner = '%s%s'%(self.bms.owner, data)
            
    def endElement(self,name):
        name = string.lower( name )
##        print 'end', name
        if name=="h3":
            self.cur_elem=None
        elif name=="dl":
            self.bms.leave_folder()
        elif name=="a" or name == 'title':
            self.cur_elem=None

# --- Test-program

if __name__ == '__main__':
    import sys

    if len(sys.argv)<2 or len(sys.argv)>3:
        print
        print "A simple utility to convert Netscape bookmarks to XBEL."
        print
        print "Usage: "        
        print "  ns_parse.py <ns-file> [<xbel-file>]"
        sys.exit(1)        
    
    ns_handler=NetscapeHandler()
    from xml.sax import make_parser
    p = make_parser(['xml.sax.drivers2.drv_sgmlop_html'])
    p.setContentHandler(ns_handler)
    file = open(sys.argv[1], 'r')
    p.parse(file)
    bms = ns_handler.bms

    if len(sys.argv)==3:
        out=open(sys.argv[2],"w")
        bms.dump_xbel(out)
        out.close()
    else:
        bms.dump_xbel()

