# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
__revision__ = '$Id: Proxy.py,v 1.3 2002/02/20 15:51:53 zoe Exp $'
from narval.lib import * # get_encoding
from narval.utils.BinaryStore import BinaryStore
import httplib
import urlparse
from urllib import quote, unquote
import htmllib
import formatter
import re
import os

try:
    lg_factory = load_extension_module('language','guess_language')
except:
    lg_factory = None

# set this variable to :
    # 0  for no junkbuster or external junkbuster
    # 1  for junkbuster based on server name
    # 2  for junkbuster based on server domain name 
    # 4  for junkbuster based on well-known patterns in spam.xml
    # 8  for data filtering junkbuster (EXPERIMENTAL !)
    # 16 for translation on the fly
    # this values can be logically ORed

MOD_XML='''<?xml version="1.0" encoding="ISO-8859-1"?><module xmlns:al="%s">'''%AL_NS

def load_spam(filePath):
    """This function is used to load the Spam-policy Tree
    once for all.
    """
    try:
        spamElement=read_domtree_from_uri(filePath)
        return spamElement.documentElement
    except IOError:
        #print sys.exc_info()[0]," unable to open : ",filePath
        return None

try:
    SPAM=load_spam("file://"+os.path.abspath(config.get_home())+"/data/spam.xml")
    BLANKPATH= os.path.abspath(config.get_home())+"/data/tr.png"
    BLANKDATA= open(BLANKPATH,"r").read()
except:
    print "Warning: spam.xml or tr.png don't exist (get_and_filter will crash)"

##
## SpamParser class
##

class SpamParser(htmllib.HTMLParser):
    """Class used to handle spam over a Web page"""
    
    def __init__(self,junkbuster_method,format):
        self.junkbuster_method=junkbuster_method
        htmllib.HTMLParser.__init__(self,format)
    
    def anchor_bgn(href,name,type):pass
    def anchor_bgn(self,href,name,type):pass
    def anchor_end(self):pass
    def save_bgn(self):pass
    def save_end(self):pass
    def handle_image(self,source,alt='',ismap='',align='',width=0,height=0):
        """ method which contains all the tests for deciding if an image
        must be trashed or not """
        
        print "SOURCE:", source
        img_server=urlparse.urlparse(source)[1]    
        print "IMAGE SERVER:", img_server
        if not img_server:
            return
    
        print "SELF.SERVER:", self.server
        # if we have to trash according to server names
        if self.junkbuster_method&1:
            if img_server!=self.server:
                self.spam.append(source)
                print "SPAM IS:", self.spam

        # if we have to trash according to domain names
        if self.junkbuster_method&2:
            if img_server[-self.domainnamelength:]!=self.domainname:
                self.spam.append(source)

    def trash_spam(self,html):

        # browse all the images'url to trash (collected while parsing page)
        # and replace them by a blank image
        for url in self.spam:
            rgxp=re.compile('[Ss][Rr][Cc][ \t]*=[ \t]*["\']?'+
                            re.escape(url)+'[ \t]*["\']?',re.M)

            imagePath = '"file://%s"'%BLANKPATH
            html=rgxp.sub('src='+imagePath,html)

        # <filter/> rules
        # experimental...
        if self.junkbuster_method&8:
            sm=SpamMatcher(self.spam_policy)
            return sm.filter(html,self.url,self.url_tuple)
        else:
            return html
        

##
##  SpamMatcher class
##

class SpamMatcher:
    def __init__(self,spampolicytree):
        self.rules={}
        self.instructions=[]
        self.filters=[]
        for rule in select_many(spampolicytree, 'rules/*'):
            self.rules[rule.getAttributeNS(NO_NS,'name')]=rule

        for foo in select_many(spampolicytree, 'policy/*'):
            if foo.nodeName in ['allow','trash']:
                self.instructions.append(foo)
                continue
            # here, we know foo.nodeName is 'filter'
            url_match=select_many(foo, 'url_match/*')

            try:
                replace_by=select_one(foo,'replace_by/text()').data
            except SelectException:
                replace_by=''
            text_match=[]
            for bar in select_many(foo,'text_match/text()'):
                text_match.append(bar.data)
            fh=foo.getAttributeNS(NO_NS,'host')
            self.filters.append((url_match,text_match,replace_by,fh))

    def test_http_rule(self, tree, url, url_tuple):
        """ Calls recursive 'eval_rules_tree' if url corresponds
            to a filtering rule."""
        if tree.getAttributeNS(NO_NS,'reverse')!='yes':
            rev=None
        else:
            rev=1
        z=tree.getAttributeNS(NO_NS,'url')
        if z and not re.match(z,url,re.VERBOSE|re.M):
            return 0
        pr,h,pat,par,q,f=url_tuple
        z=tree.getAttributeNS(NO_NS,'host')
        if z and not re.match(z,h,re.VERBOSE|re.M):
            return 0
        z=tree.getAttributeNS(NO_NS,'path')
        if z and not re.match(z,pat,re.VERBOSE|re.M):
            return 0
        z=tree.getAttributeNS(NO_NS,'protocol')
        if z and not re.match(z,pr,re.VERBOSE|re.M):
            return 0
        z=tree.getAttributeNS(NO_NS,'param')
        if z and not re.match(z,par,re.VERBOSE|re.M):
            return 0
        z=tree.getAttributeNS(NO_NS,'query')
        if z and not re.match(z,q,re.VERBOSE|re.M):
            return 0
        z=tree.getAttributeNS(NO_NS,'fragid')
        if z and not re.match(z,f,re.VERBOSE|re.M):
            return 0
        print "TESTS SUCCESSFULLY PASSED"
#        ??????????????? WHAT'S THAAAAT???
#        return 1
        children=select_many(tree,'*')
        if not len(children):
            return not rev
        for foo in children:
            if self.eval_rules_tree(foo, url, url_tuple):
                return not rev
        return rev

    def eval_rules_tree(self,tree,url,url_tuple):
        """This complex recursive method checks if
           url belong to a filtering rule, and 
           recursively calls if have to check path
           deeply.
        """
        t=tree.nodeName
        if t=='rule':
            return self.test_http_rule(tree,url,url_tuple)
        if tree.getAttributeNS(NO_NS,'reverse')!='yes':
            rev=0
        else:
            rev=1
        if t=='ruleref':
            if not self.test_http_rule(self.rules[tree.getAttributeNS(NO_NS,'rule')],url,url_tuple):
                return rev
            children=select_many(tree,'*')
            if not len(children):
                return rev^1
            else:
                for foo in children:
                    if self.eval_rules_tree(foo,url,url_tuple):
                        return rev^1
                return rev
        if t=='or':
            for rule in select_many(tree,'*'):
                if self.eval_rules_tree(rule,url,url_tuple):
                    return rev^1
            return rev
        # here, we know t=='and'
        for rule in select_many(tree,'*'):
            if not self.eval_rules_tree(rule,url,url_tuple):
                return rev
            return rev^1

    # this returns 1 if the url should be blocked
    def eval_url(self, url, url_tuple):
        """Calls recursive eval_rules_tree"""
        print "EVAL URL"
        for instruction in self.instructions:
            for rules_tree in select_many(instruction,'*'):
                
                if self.eval_rules_tree(rules_tree,url,url_tuple):
                    print "OK RULE TREES"
                    if instruction.nodeName=='allow':
                        return 0
                   # here, we know instruction.nodeName=='trash'
                    return 1
        return 0

    # this takes a text, applies filters on it, and returns the filtered text
    def filter(self, text, url, url_tuple):
        print "CALLING SPAM FILTER"
        for url_match,text_match,replace_by,fh in self.filters:
            if fh:
                print "PATTERN OBJECT", fh, "url:", url_tuple
                if not re.match(fh,url_tuple[1],re.VERBOSE|re.M):
                    continue
                else:
                    print "Bonjour"
            for foo in url_match:
                print "EVAL RULES TREE", PrettyPrint(foo)
                if self.eval_rules_tree(foo,url,url_tuple):
                    for bar in text_match:
                        print "TEXT=", text
                        rgxp=re.compile(bar,re.VERBOSE|re.M)
                        text=rgxp.sub(replace_by,text)
        return text

##
## gettype : returns the type contained in the Contain-Type field of a string
##           of MIME headers
##

def gettype(headers):
    for line in string.split(headers,'\n'):
        if string.lower(line[:12]) == 'content-type':
            beg=13
            while line[beg]==' ' or line[beg]=='\t':
                beg=beg+1
            end=beg+1
            while line[end] not in ['\n',';','\r']:
                end=end+1
            return line[beg:end]
    return ''


def domainname(dns):
    """
    returns the domain name part of a DNS
    """
    dot=string.find(dns,'.')
    if dot in [-1,0]:
        return ''
    return dns[dot+1:]

## ---------------------------------------------------------------------------
## NOTE URL
## ---------------------------------------------------------------------------
def note_url_f(args, doc) :
    import time
    a = args['input']
    url=select_one(a,'url/text()').data
#    ref=select_many(a,'headers/header/text()')

    f = open(config.get_home()+'/data/urlvisited.xml','a+')
    f.write('<url url="'+re.sub('<','&lt;',re.sub('&','&amp;',url))+'" ')
    f.write('date="'+time.asctime(time.localtime(time.time()))+'"/>\n')
    return {}

MOD_XML=MOD_XML+''' 
<al:action name="note_url" func="note_url_f">
  <al:description lang="en">Add the url to the list of visited urls</al:description>
   <al:description lang="fr">Ajoute l\'url  la liste des url visites</al:description>
  <al:input id='input'><al:match>http-request</al:match>
  <al:description lang="en">request matching url to add</al:description></al:input>
</al:action>'''  

## ---------------------------------------------------------------------------
## FUNCTIONS HELPERS FOR GET_AND_FILTER
## ---------------------------------------------------------------------------
def get_replace(obj):
    """ replace special characters ( %3F --> ?)"""
    str = re.split('&url=', obj.group())
    return 'href="%s"'%unquote(str[-1])

def getHeaderType(headers, type_name):
    """ get the header element which attribute is 'Content-Type'
        content-type can match */*; charset=* regexp
    """
    import re
    for h in headers:
        name = h.getAttributeNS(NO_NS,'name')
        if name.lower()==type_name.lower():
            try:
                content_type= select_one(h, 'text()').data
                mat = re.match(re.compile("([\/\w]*);\s*charset=(.*)"), content_type)
                if mat:
                    return (mat.group(1).strip(), mat.group(2).strip())
                # if no charset specified/ CAUTION: image type has no charset
                return (content_type, "iso-8859-1")
            except:
                print "Warning: HTTP-request hasn't charset specification,\
default will be iso-8858-1"
                return ("text/html", "iso-8859-1")

def filter(response, doc, junkbuster_method):
    """Filters images and replace text"""
#    ElementType = select_one(response,'headers/header[@name="Content-Type"]')
    textElementList = select_many(response,'headers/header')
    if getHeaderType(textElementList, 'content-type')[0]=='text/html':
        # if page to filter is of text/html type,
        # so we parse it to suppress images at 'src' location

        bar=select_one(response,'content/text()')
        
        sp=SpamParser(junkbuster_method,formatter.NullFormatter())

        sp.url=select_one(response,'url/text()').data
        sp.url_tuple=urlparse.urlparse(sp.url)
        # netlocation server=www.baboul.com or domainname=baboul.com
        sp.server=sp.url_tuple[1]
        sp.domainname=domainname(sp.server)
        sp.domainnamelength=len(sp.domainname)
        
        sp.spam=[]
#       sp.spam_policy = args['input_spam']blabla or SPAM
        sp.spam_policy=SPAM

        # handle_image: add src to spam if has to be trashed
        sp.handle_image(sp.url)
        sp.feed(bar.data)
        bar.data=sp.trash_spam(bar.data)
       
        pattern = re.compile('href[\t  ]*=[\t ]*["\']?[^\\?>]+\\?lp[^&]*&url=(?P<adr>[^ \t>"\']*)["\']?',re.I)
        bar.data = re.sub(pattern,get_replace,bar.data)

    return response

def findDestLanguage(spokenList,destList):
    """ """
    for l in spokenList:
        try:
            index = destList.index(l)
            return destList[index]
        except ValueError:
            pass

# this function takes in arguments the "source language" (Web page's language),
# the list of spoken languages, and returns the translation type ('en_fr' for instance)
def choose_translation_type(sourceLanguage,spokenList):

    # these lists are made according to Altavista possibilities
    if sourceLanguage == 'en':
        destList = ['fr','de','it','es','pt']
    elif sourceLanguage == 'de':
        destList = ['en','fr']
    elif sourceLanguage == 'fr':
        destList = ['en','de']
    elif sourceLanguage == 'it':
        destList = ['en']
    elif sourceLanguage == 'es':
        destList = ['en']
    elif sourceLanguage == 'pt':
        destList = ['en']

    destLanguage = findDestLanguage(spokenList,destList)
    if destLanguage is None:
        return None

    return sourceLanguage+'_'+destLanguage

###############################################################
#         special characters (and translation) list
#
#  " " : "<" : ">" : "#" : "%" : "{" : "}" : "|"
#  "\\" : "^" : "~" : "[" : "]" : "`" : ";"
#  "/" : "?" : ":" : "@" : "=" : "&" : "$";
#  "%20" : "%3C" : "%3E" : "%23" : "%25"
#  "%7B" : "%7D" : "%7C" : "%5C" : "%5E
#  "%2F" : "%3F" : "%3A" : "%40" : "%3D"
#  "%26" : "%24" )
###############################################################

def build_redirected_url(proxyString,lang,url):
    url = quote(url)
    return proxyString+'lp='+lang+'&url='+url
    #http://www.lycos.fr/cgi-bin/pursuit?matchmode=and&mtemp=main&etemp=error&query=astronomie&cat=fr

def replace_url(url):
    list = string.split(url,'?')
    base = list.pop(0)
    # if there is more than one '?' in the original address
    right_part = string.join(list,'?')
    
    if len(right_part) > 0:
        # get the "spec_list"
        spec_list = string.split(right_part,'&')
        # remove "BabelFishForm=GET" if needed
        if right_part == 'BabelFishForm=GET':
            spec_list.pop(0)
            right_part = string.join(spec_list,'&')

        # retrieve the "url" part
        list = string.split(right_part,'url=')
        right_list = list[1:]
                
        # if 'url=' appear in the "real" address; replace special characters
        right_part = quote(string.join(right_list,'url='))
        # right_part = string.join(spec_list)
        
        if len(right_part) > 0:
            right_part = list[0]+'url='+right_part
        else:
            right_part = list[0]

        url = base+'?'+right_part

    return url

def get_and_filter_f(args, doc):
    """ function which retrieves a parsed request, makes it,
    and filter according to the junkbuster_method's value
    """
    # retrieve the request
    event = args['input_http']
    datas = select_many(event, "data/text()")
    # if special characters, replace them
    url=replace_url(select_one(event,'url/text()').data)
    uid = event.getAttributeNS(NO_NS,'unique-id')

    # urlparse format : scheme://netloc/path;parameters?query#fragment
    ask_url=urlparse.urlparse(url)
    netloc=ask_url[1]
    # prepare output 
    h, ne, nc, nh = prepare_output(doc, uid, url)

    # find which type of junkbuster is on
    jbm = args['input_junkbuster']
    jbmvalue = jbm.getAttributeNS(NO_NS,'value')
    if jbmvalue: junkbuster_method= int(jbmvalue)
    else: junkbuster_method=None
    
    # if junkbuster is on, and is filter-image case with spam.xml file
    # during the loading of webpage
    if junkbuster_method & 4:
        sm=SpamMatcher(SPAM)
        if sm.eval_url(url, ask_url):
            headers  = 'Content-Length: %s\r\n'%len(BLANKDATA)
            headers += 'Connection: close\r\nContent-Type: image/png\r\n'
            return complete_response(h, ne, nc, nh, doc, 200, 'OK', headers,
            BLANKDATA, junkbuster_method, ('url','"file://%s"'%BLANKPATH))

    # send the request and use a proxy if asked in arguments
    proxy = args['input_proxy']
    zh = _proxy_test(event, url, ask_url, proxy, 1)
    if not zh: 
        return { 'output': h }

    # translate request headers <header> nodes in a tuples list and get reply
    tuples_list = h_nodes2tuples(select_many(event, "headers/header"))
    errcode, errmsg, headers = get_reply(zh, tuples_list, datas)
    link = None     # this is the object that references downloaded images

    # if no problems, and junkbuster is not 4
    if errcode == 200:

        f=zh.getfile()
        _type = gettype(str(headers))
        data = f.read()
        # if not an image, try to find the language and translate it if needed
        if _type == 'text/html':
            if lg_factory and junkbuster_method&16:

                spokenElement = args['input_spoken']
                if spokenElement :

                    languageList=spokenElement.childNodes
                    # create a list containing all the spoken languages
                    spokenList = []
                    for languageElement in languageList:
                        spokenList.append(languageElement.getAttributeNS(NO_NS,'code'))
                    # determine the page's language
                    # sourceLanguage = lg_factory.identify_language_from_url(url)
                    sourceLanguage = lg_factory.identify_language_from_buffer(data)

                    #if not spoken
                    if not sourceLanguage in spokenList:
                        lang = choose_translation_type(sourceLanguage,spokenList)

                        if lang:
                            url = build_redirected_url('http://babel.altavista.com/urltrurl?',lang,url)
                            #url = build_redirected_url('http://fr.altavista.com/scripts/eusearchengine.dll?',lang,url)
                            ask_url=urlparse.urlparse(url)
                            # now retrieve the new page's data
                            zh = _proxy_test(event, url, ask_url, proxy)
                            if not zh : 
                                return {'output': h}
                            # get server reply
                            errcode, errmsg, headers = get_reply(zh, tuples_list, datas, ask_url[1])
                            if errcode != 200:
                                data = error_data(str(errcode), errmsg, str(headers))
                            # if no error, get data, which translated text
                            else:
                                f = zh.getfile()
                                data = f.read()

        elif _type[:5]=="image" :#and junkbuster&~4
            link = ('url', url)
        f.close()
    # if an error occured
    else:
        data = error_data(str(errcode),errmsg,str(headers))
    
    return complete_response(h, ne, nc, nh, doc, errcode, errmsg, headers, 
                             data, junkbuster_method, link)


MOD_XML=MOD_XML+'''
<al:action name="get_and_filter" func="get_and_filter_f">
  <al:description lang="en">Filter out web junk (ads, cookies, etc.)</al:description>
  <al:description lang="fr">Filtre les ordures du web (pubs, cookies, etc.)</al:description>
  <al:input id='input_http'><al:match>http-request[url/text()]</al:match>
  <al:description lang="en">http request to filter</al:description></al:input>
  <al:input id='input_proxy' optional="yes"><al:match>proxy[@type="http"]</al:match>
  <al:description lang="en">proxy element, for page requests</al:description></al:input>
  <al:input id='input_spam' optional="yes"><al:match>spam-policy</al:match>
  <al:description lang="en">spam policy element</al:description></al:input>
  <al:input id='input_junkbuster' optional="yes"><al:match>junkbuster-method</al:match>
  <al:description lang="en">junkbuster configuration element</al:description></al:input>
  <al:input id='input_spoken' optional="yes"><al:match>spoken-languages</al:match>
  <al:description lang="en">node containing user\'s spoken languages</al:description></al:input>
  <al:output id='output'><al:match>http-response</al:match>
  <al:description lang="en"></al:description></al:output>
</al:action>'''

## ---------------------------------------------------------------------------
## FUNCTIONS HELPERS FOR GET_AND_FILTER (SUITE)
## ---------------------------------------------------------------------------
def error_data(errcode, errmsg, headers):
    """ """
    return '<p>The server returned the following error : </p><p><strong>%s \
              - %s</strong></p><p>And gave the following headers : </p>\
              <p><pre>%s</pre></p>'%(errcode, errmsg, headers)

def _encode(data, encoding="utf-8"):
    """encode data in unicode format"""
    try:
        return unicode(data, encoding)
    except UnicodeError:
        try:
            return unicode(data, "ISO-8859-1")
        except UnicodeError:
            print "UNICODE ERROR : no valid encoding has been specified"
            return data

def export_http_proxy(proxy):
    """export http proxy..."""
    from os import environ
    environ['http_proxy'] = 'http://%s:%s' % (proxy.getAttributeNS(NO_NS,'hostname'),
                                              proxy.getAttributeNS(NO_NS,'port'))

def h_nodes2tuples(h_nodes):
    """ tranforms headers/headers nodes in list of tuples"""
    return [
        (h.getAttributeNS(NO_NS,'name'), select_one(h, 'text()').data)
        for h in h_nodes ]

def str_h2h_nodes(str_headers, doc):
    """ tranforms text headers to headers/headers nodes, in a list"""
    list = []
    for a in string.split(str_headers,'\015\012'):
        try:
            (k,v)=tuple(string.split(a,': '))
            h= doc.createElementNS(NO_NS,'header')
            h.setAttributeNS(NO_NS,'name',k)
            h.appendChild(doc.createTextNode(v))
            list.append(h)
        except ValueError:
            pass
    return list

def _proxy_test(event, url, ask_url, proxy, export=None):
    """ send to server the request defined by url, in case of proxy
        "manually" specified, or client definition ?
        Returns None if sending the request wasn't a success.
    """
    if proxy:
        print "PROXY"
        if export:
            export_http_proxy(proxy)
        zh=httplib.HTTP(proxy.getAttributeNS(NO_NS,'hostname'),
                        int(proxy.getAttributeNS(NO_NS,'port')))
        zh.putrequest(string.upper(event.getAttributeNS(NO_NS,'type')),url)
        return zh
    else:
        try:
            zh = httplib.HTTP(ask_url[1])
            ask_url =\
            urlparse.urlunparse( ('', '', ask_url[2], ask_url[3],
                                ask_url[4], ask_url[5]) )
            zh.putrequest(string.upper(event.getAttributeNS(NO_NS,'type')), ask_url)
            return zh
        except :
            print sys.exc_info()[0],\
            ' -> impossible to connect to "',ask_url[1],'"'
            return None

def prepare_output(doc, uid, url):
    """ prepare the http-response """
    h = doc.createElementNS(NO_NS,'http-response')
    h.setAttributeNS(NO_NS,'unique-id', uid)

    nu = doc.createElementNS(NO_NS,'url')
    nu.appendChild(doc.createTextNode(url))
    h.appendChild(nu)
    ne = doc.createElementNS(NO_NS,'errmsg')
    h.appendChild(ne)
    nh = doc.createElementNS(NO_NS,'headers')
    h.appendChild(nh)
    nc = doc.createElementNS(NO_NS,'content')
    h.appendChild(nc)
    return (h, ne, nc, nh)

def complete_response(response, ne, nc, nh, doc, errcode, errmsg, headers, data,
                        junkbuster_method, link):
    """ complete the http-response
        link is a tuple, link[0]=type of data, link[1]=data itself (url or id)
    """
    response.setAttributeNS(NO_NS,'status',str(errcode))
    ne.appendChild(doc.createTextNode(errmsg))
    hlist = str_h2h_nodes(str(headers), doc)
    for _h in hlist: nh.appendChild(_h)

    _type, charset = getHeaderType(hlist, "content-type")
    
    if _type[:4]=="text":
        nc.appendChild(doc.createTextNode(_encode(data, charset)))
        if junkbuster_method&~4:
            response = filter(response, doc, junkbuster_method)
    elif link:
        image_ref = doc.createElementNS(NO_NS,'reference')
        image_ref.setAttributeNS(NO_NS,'type',link[0])
        image_ref.appendChild(doc.createTextNode(link[1]))

        if link[1] not in get_binary_store().get_urls():
            get_binary_store().add_binary(link[1], data)

        response.appendChild(image_ref)
        
    return {'output':response}

def get_reply(zh, tuples_list, datas, netloc=None):
    """ Get the server reply """
    for (k,v) in tuples_list:
        if netloc and k == 'host' : v = netloc
        zh.putheader(k,v)
    # send a blank line
    zh.endheaders()
    # send more data if needed by the request
    for foo in datas:
        zh.send(foo.data)
    # get reply
    return zh.getreply()


MOD_XML=MOD_XML+'</module>'

if __name__ == "__main__" :
    print MOD_XML
    from xml.dom.ext.reader import Sax2
    doc=Sax2.FromXml(MOD_XML)
    print doc.documentElement
